/**
 * \file
 * \brief  Implementation of the window with the least squares method
 * results.
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#ifndef __GRAPH_H
#define __GRAPH_H

enum GraphWindowFitMode	{
	FIT_VALS,
	FIT_VALNTERMS_INITAPPROX,
	FIT_VALNTERMS_AUTO,
	FIT_EXTERMS_ONLY,
	FIT_NOOPT
};


/**
 * \brief  Panel (set of controls) that shows deviations for H or Cp data sets
 *
 * Used by GraphWindow class
 */
class DevInfoPanel : public wxPanel
{
public:
	DevInfoPanel(wxWindow *parent, wxPoint pos, wxSize size, wxString unit = wxString(L"J / (mol\u00B7K)"));
	void LoadLSQDevInfo(LSQDevInfo &dev);
	
private:
	wxTextCtrl *txtMaxAbsDev, *txtMaxAbsDevX;
	wxTextCtrl *txtMaxRelDev, *txtMaxRelDevX;
	wxTextCtrl *txtStdAbsDev;
	wxTextCtrl *txtStdRelDev;
};

/**
 * \brief  Panel (set of controls) that shows deviations for data sets individually
 */

class DevInfoListPanel : public wxPanel
{
public:
	DevInfoListPanel(wxWindow *parent, wxPoint pos, wxSize size);
	void LoadLSQDevsInfo(LSQResult &result);
	void LoadLSQDevsInfo(std::vector <LSQDevInfo> &devs);
	void OnSize(wxSizeEvent &event);
	void OnCmbPointsFilter(wxCommandEvent &event);
private:
	wxComboBox *cbPointsFilter;
	wxGrid *grDevs;
	std::vector <LSQDevInfo> devs;
	std::vector <LSQDevInfo> devsBaseLine;
	std::vector <LSQDevInfo> devsAnomaly;
	wxDECLARE_EVENT_TABLE();
};



/**
 * \brief  Window with the least squares method results.
 *
 * It shows the next results of the least squares method:
 * * Optimized values of alpha and theta coefficients
 * * Optimized values of extra terms parameters (in the form of text report)
 * * Deviations for Cp data (absolute and relative deviations, maximal deviatons)
 * * Tools for management of plots and tables
 */
class GraphWindow : public wxFrame
{
public:
	GraphWindow(MainWindow *parent, const wxString &title);
	void OnSize(wxSizeEvent &event);
	void OnClose(wxCloseEvent &event);
	void AddData(LSQContext lsq);
	void AddParam0(std::vector <double> alpha, std::vector <double> theta);
	void CurveFit(GraphWindowFitMode fitMode = FIT_VALS);
private:
	LSQContext lsq;
	LSQResult lsqRes;
	std::vector <double> alpha0;
	std::vector <double> theta0;
	
	wxGrid *grPar;
	wxButton *btnCopyParams;
	
	wxStaticBox *boxParams, *boxLsq;
	wxPanel *ctlBlParams;
	wxPanel *ctlExParams;
	wxNotebook *nbkParams;
	wxTextCtrl *txtExParamsReport;
	
	wxNotebook *nbkTabsDev;	
	DevInfoPanel *CpDevPanel;
	DevInfoPanel *HDevPanel;
	DevInfoListPanel *SeriesDevPanel;

	void OnCopyParams(wxCommandEvent &event);
	void OnShowTdTable(wxCommandEvent &event);
	void OnShowCpResTable(wxCommandEvent &event);
	void OnShowHResTable(wxCommandEvent &event);
	void OnCpPlot(wxCommandEvent &event);
	void OnCpTPlot(wxCommandEvent &event);
	void OnHPlot(wxCommandEvent &event);
	void OnSPlot(wxCommandEvent &event);
	void OnDHPlot(wxCommandEvent &event);

	void OnCpAbsResPlot(wxCommandEvent &event);
	void OnCpRelResPlot(wxCommandEvent &event);

	void OnHAbsResPlot(wxCommandEvent &event);
	void OnHRelResPlot(wxCommandEvent &event);

	void OnSaveAsCSV(wxCommandEvent &event);
	void OnSaveAsHTML(wxCommandEvent &event);

	void OnDebyeCorrection(wxCommandEvent &event);

	wxDECLARE_EVENT_TABLE();	
};
#endif
