/**
 * \file
 * \brief Tests for TDPEFunc class
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <math.h>
#include <vector>
#include <algorithm>
#include <functional>
#include <string>
#include <stdio.h>
#include "datafit.h"


/**
 * \brief Shows a table of tabulated thermodynamic values using optimized parameters
 *
 * This test reproduces the table of tabulated thermodynamic values from
 * http://dx.doi.org/10.1007/s00269-013-0625-1. Note that currently its validity
 * must be checked manually by a user!
 */
void test1()
{
	// Show the legend about the test
	printf("TEST: REPRODUCING THE TABLE OF TABULATED VALUES\n");
	printf("SOURCE OF PARAMETERS AND DATA: http://dx.doi.org/10.1007/s00269-013-0625-1\n");
	printf("  S.V. Kurdakova, R.O. Grishchenko, A.I. Druzhinina, L.P.Ogorodova\n");
	printf("  Thermodynamic properties of synthetic calcium-free carbonate cancrinite //\n");
	printf("  // Physics and Chemistry of Minerals. 2014. 41(1). P. 75-83\n");
	// Create object using existing parameters
	double param[] = {0.9778,49.185, 12.36,164.19, 20.96,404.9, 30.3,1095};
	TDPEFunc tdf(param, 8);
	// Create T vector
	std::vector <double> T;	
	for (double Ti = 5.0; Ti < 81.0; Ti += 5.0) {
		T.push_back(Ti);
	}
	for (double Ti = 90; Ti <= 301; Ti += 10.0) {
		T.push_back(Ti);
	}
	T.push_back(273.15);
	T.push_back(298.15);
	std::sort(T.begin(), T.end());
	// Calculate Cp, S and H vectors for the table	
	std::vector <double> Cp = tdf.CpFunc(T);
	std::vector <double> S = tdf.SFunc(T);
	std::vector <double> H = tdf.HFunc(T);
	
	// Show the table	
	printf("%5s %10s %10s %10s\n", "T", "Cp", "S", "H");
	for (int i = 0; i < T.size(); i++) {
		printf("%6.2f %10.2f %10.2f %10.4f\n", T[i], Cp[i], S[i], H[i]/1000.0);
	}
	printf("\n\n");
	// Final print
	printf("\n\n");
}

/**
 * \brief Tests Cp(T), S(T)-S(0) and H(T)-H(0) derivatives
 *
 * This tests checks validity of Cp(T), S(T)-S(0) and H(T)-H(0) derivatives
 * using numerical methods. Note that currently its validity must be checked
 * manually by a user!
 */
void test2()
{
	printf("TEST: TESTING CP, S and H DERIVATIVES\n");
	// Create object using existing parameters
	double param[] = {0.9778,49.185, 12.36,164.19, 20.96,404.9, 30.3,1095};
	double param_tmp[8];
	TDPEFunc tdf(param, 8);
	// Allocate memory for derivatives calculations	
	int nparams = tdf.GetNumOfParams();
	std::vector <double> JxCp(nparams);
	std::vector <double> JxS(nparams);
	std::vector <double> JxH(nparams);
	std::vector <double> JxCpNum(nparams);
	std::vector <double> JxSNum(nparams);
	std::vector <double> JxHNum(nparams);
	double T = 298.15;
	double dB = 1e-7;
	// Calculate derivatives using built-in functions
	tdf.dCpdBFunc(T, &JxCp[0]);
	tdf.dSdBFunc(T, &JxS[0]);
	tdf.dHdBFunc(T, &JxH[0]);
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxCp[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxS[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxH[i]);
	}
	printf("\n\n");


	// Calculate derivatives using numerical methods
	for (int i = 0; i < 8; i++) {
		std::copy(param, param + 8, param_tmp);
		param_tmp[i] += dB;
		TDPEFunc tdf2(param_tmp, 8);
		double Cp2 = tdf2.CpFunc(T);
		double S2  = tdf2.SFunc(T);
		double H2  = tdf2.HFunc(T);
	
		std::copy(param, param + 8, param_tmp);
		param_tmp[i] -= dB;
		TDPEFunc tdf1(param_tmp, 8);
		double Cp1 = tdf1.CpFunc(T);
		double S1  = tdf1.SFunc(T);
		double H1  = tdf1.HFunc(T);

		JxCpNum[i] = (Cp2 - Cp1) / (2*dB);
		JxSNum[i] = (S2 - S1) / (2*dB);
		JxHNum[i] = (H2 - H1) / (2*dB);
	}
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxCpNum[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxSNum[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxHNum[i]);
	}
	printf("\n\n");
	// Final print
	printf("\n\n");		
}

/**
 * \brief Runs two tests
 */
int main()
{
	test1();
	test2();
	
	return 0;
}
