/**
 * \file
 * \copyright (C) 2016-2017 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <algorithm>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/notebook.h"
#include "datafit.h"
#include "plotcanvas.h"
#include "cp.h"
#include "graph.h"
#include "xyplots.h"
#include "misc.h"

enum {
	ID_MNU_SAVEAS = 100,
	ID_MNU_EDIT,
	ID_MNU_COPY,
	ID_MNU_CHKLEGEND,
	ID_MNU_CHKGRID
};

wxBEGIN_EVENT_TABLE(XYPlotWindow, wxFrame)
	EVT_SIZE(XYPlotWindow::OnSize)
	EVT_MENU(ID_MNU_SAVEAS, XYPlotWindow::OnSaveAs)
	EVT_MENU(ID_MNU_COPY, XYPlotWindow::OnCopy)
	EVT_MENU(ID_MNU_CHKLEGEND, XYPlotWindow::OnChkLegend)
	EVT_MENU(ID_MNU_CHKGRID, XYPlotWindow::OnChkGrid)
wxEND_EVENT_TABLE()


/**
 * \brief Initializes controls and saves user-defined data
 */
XYPlotWindow::XYPlotWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes)
	: wxFrame(parent, wxID_ANY, title, wxPoint(10, 10), wxSize(400, 400))
{
	this->lsqCont = lsqCont;
	this->lsqRes = lsqRes;
	wxPanel *panel = new wxPanel(this, -1);
	pl = new PlotCanvas(panel, -1, wxPoint(50, 50), wxSize(300, 300));
	SetIcon(wxICON(sample));
	// Create user menu	
	menuBar = new wxMenuBar();
	wxMenu *menuFile = new wxMenu(); // File
	menuFile->Append(ID_MNU_SAVEAS, "Save as", "Save the plot as...");
	menuBar->Append(menuFile, "&File");
	wxMenu *menuEdit = new wxMenu(); // Edit
	menuEdit->Append(ID_MNU_COPY, "Copy", "Copy the plot to the clipboard");
	menuEdit->AppendSeparator();
	menuEdit->Append(ID_MNU_CHKLEGEND, "Legend", "Show/hide legend", wxITEM_CHECK);
	menuEdit->FindItem(ID_MNU_CHKLEGEND)->Check(true);
	menuEdit->Append(ID_MNU_CHKGRID, "Grid", "Show/hide grid", wxITEM_CHECK);
	menuEdit->FindItem(ID_MNU_CHKGRID)->Check(true);
	menuBar->Append(menuEdit, "&Edit");	
	SetMenuBar(menuBar); // Append the whole menu to the form
	// Create some default markers
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,255), wxColor(0,255,255), LINE_NONE, MARKER_CIRCLE));
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,255), wxColor(0,255,255), LINE_NONE, MARKER_SQUARE));
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,0),   wxColor(255,255,255), LINE_NONE, MARKER_CIRCLE));
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,0),   wxColor(255,255,255), LINE_NONE, MARKER_SQUARE));
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,0),   wxColor(255,255,0), LINE_NONE, MARKER_CIRCLE));
	seriesMarkers.push_back(PlotMarker(wxColor(0,0,0),   wxColor(255,255,0), LINE_NONE, MARKER_SQUARE));
	// Set window coordinates
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetMinSize(wxSize(250, 250));
}

/**
 * \brief To be redefined by user
 */
void XYPlotWindow::InitPlot()
{
}

/**
 * \brief Resizes PlotCanvas inside the window
 */
void XYPlotWindow::OnSize(wxSizeEvent &event)
{
        wxSize s = this->GetClientSize();
	s.SetWidth(s.GetWidth() - 10);
	s.SetHeight(s.GetHeight() - 10);
	pl->SetSize(s);
	pl->Move(5, 5);
	event.Skip();
}

void XYPlotWindow::OnChkLegend(wxCommandEvent &event)
{
	wxMenu *mnu = (wxMenu *) event.GetEventObject();
	wxMenuItem *item = mnu->FindItem(ID_MNU_CHKLEGEND);
	bool isChecked = item->IsChecked();
	if (isChecked) {
		pl->EnableLegend();
	} else {
		pl->DisableLegend();
	}
}

void XYPlotWindow::OnChkGrid(wxCommandEvent &event)
{
	wxMenu *mnu = (wxMenu *) event.GetEventObject();
	wxMenuItem *item = mnu->FindItem(ID_MNU_CHKGRID);
	bool isChecked = item->IsChecked();
	if (isChecked) {
		pl->EnableGrid();
	} else {
		pl->DisableGrid();
	}
}

/**
 * \brief Enables legend at the 2D plot and checks corresponding menu item
 */
void XYPlotWindow::EnableLegend()
{
	menuBar->FindItem(ID_MNU_CHKLEGEND)->Check(true);
	pl->EnableLegend();	
}

/**
 * \brief Disables legend at the 2D plot and unchecks corresponding menu item
 */
void XYPlotWindow::DisableLegend()
{
	menuBar->FindItem(ID_MNU_CHKLEGEND)->Check(false);
	pl->DisableLegend();
}


/**
 * \brief Saves 2D plot into file
 *
 * Currently only .png files (with screen resolution) are supported
 */
void XYPlotWindow::OnSaveAs(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Save image"), "", "",
                       "png files (*.png)|*.png|wmf files (*.wmf)|*.wmf|MATLAB M-files (*.m)|*.m", wxFD_SAVE);
//                       "png files (*.png)|*.png|svg files (*.svg)|*.svg|*.ps files (*.ps)|*.ps", wxFD_SAVE);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		bool isOk = false;
		switch (openFileDialog.GetFilterIndex()) {		
		case 0: // PNG
			isOk = pl->SaveToBitmap(fileName, wxBITMAP_TYPE_PNG);
			break;
		case 1: // WMF (WinNT only!)
			isOk = pl->SaveToWMF(fileName);
			break;
		case 2: // MATLAB M files
			isOk = pl->SaveToM(fileName);
			break;
		case 3: // SVG
			isOk = pl->SaveToSVG(fileName);
			break;
		case 4: // PS
			isOk = pl->SaveToPS(fileName);
			break;
		}
		if (!isOk) {
			wxMessageBox("Cannot save the file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
	
}

/**
 * \brief  Copies data to clipboard (WIN32-specific)
 */
void XYPlotWindow::OnCopy(wxCommandEvent &event)
{
	pl->ToClipboard();
}


/**
 * \brief Initialize T-Y data window (canvas and controls)
 * \param parent Parent window
 * \param title Window title
 * \param lsqCont LSQContext class example (with user data)
 * \param lsqRes LSQResult class example (with user model)
 * \param dataType Selects data type that will be used on graph
 */
YDataWindow::YDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes, TDDataType dataType)
	: XYPlotWindow(parent, title, lsqCont, lsqRes)
{
	this->dataType = dataType;
	InitPlot();
}

/**
 * \brief  Initialize axes labels, sizes and calculate curves
 */
void YDataWindow::InitPlot()
{
	// Initialize axes labels
	pl->SetXLabel("T / K");
	switch (dataType) {
	case DATATYPE_CP:
		pl->SetYLabel(wxT("Cp / J/(mol\u00B7K)")); // \cdot
		break;
	case DATATYPE_H:
		pl->SetYLabel(wxT("H(T)-H(0) / kJ/mol"));
		break;
	case DATATYPE_S:
		pl->SetYLabel(wxT("S(T)-S(0) / J/(mol\u00B7K)"));
		DisableLegend(); // There are no markers at this plot
		break;
	default: // TODO: SOME ERROR MESSAGE
		break;
	}
	// Draw data series
	std::vector <std::string> serIDList = lsqCont.GetSeriesIDList(false, dataType);
	pl->SetLegend(serIDList, seriesMarkers, seriesMarkers[0]);
	unsigned int numOfMarkers = seriesMarkers.size();
	for (unsigned int i = 0; i < serIDList.size(); i++) {	
		int markerIndex = (i >= numOfMarkers) ? (numOfMarkers - 1) : i;
		std::vector <double> x = lsqCont.GetXVector(dataType, serIDList[i]);
		std::vector <double> y = lsqCont.GetYVector(dataType, serIDList[i]);
		if (dataType == DATATYPE_H) {
			for (unsigned int j = 0; j < y.size(); j++) {
				y[j] /= 1000.0; // J to kJ
			}
		}
		if (dataType != DATATYPE_H) {
			pl->Plot(x, y, seriesMarkers[markerIndex]);
		}
	}
	// Draw theoretical curve
	// Add to graph
	std::vector <double> x = lsqCont.GetXVector(dataType);
	std::vector <double> y = lsqCont.GetYVector(dataType);
	std::vector <double> T;
	double minT = 1, maxT = 990;
	if (x.size() >= 2) {
		if (dataType != DATATYPE_H) {
			minT = *std::min_element(x.begin(), x.end());
			maxT = *std::max_element(x.begin(), x.end());
		} else {
			std::vector <double> x1 = lsqCont.GetX1Vector(dataType);
			std::vector <double> x2 = lsqCont.GetX2Vector(dataType);
			minT = std::min(
				*std::min_element(x.begin(), x.end()),
				std::min(
					*std::min_element(x1.begin(), x1.end()),
					*std::min_element(x2.begin(), x2.end())
				)
			);
			maxT = std::max(
				*std::max_element(x.begin(), x.end()),
				std::max(
					*std::max_element(x1.begin(), x1.end()),
					*std::max_element(x2.begin(), x2.end())
				)
			);
		}
	}
	if (minT > 0.01) {
		minT = 0.01;
	}

	for (double Ti = minT; Ti <= maxT + 1e-10; Ti += (maxT - minT) * 0.001) {
		T.push_back(Ti);
	}
	// Add points from data series (to show narrow transitions better)
	if (dataType == DATATYPE_CP) {
		for (unsigned int i = 0; i < x.size(); i++) {
			T.push_back(x[i]);
		}
		std::sort(T.begin(), T.end());
	}
	// Calculate and plot Y vector
	std::vector <double> Y = lsqRes.YFunc(dataType, T);
	if (dataType == DATATYPE_H) {
		for (unsigned int i = 0; i < Y.size(); i++) {
			Y[i] /= 1000.0; // J to kJ
		}
	}
	pl->Plot(T, Y, PlotMarker(wxColor(0,0,0), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	// Enable/disable legend
	if (dataType == DATATYPE_CP) {
		EnableLegend(); // Cp
	} else {
		DisableLegend(); // There are no markers at this plot (H or S)
	}

	// Initialize axes
	pl->FitAxes();
}

DHDataWindow::DHDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes, double T0, double DT0, bool isNormed)
	: XYPlotWindow(parent, title, lsqCont, lsqRes)
{
	this->T0 = T0;
	this->DT0 = DT0;
	this->isNormed = isNormed;
	InitPlot();
}

void DHDataWindow::InitPlot()
{
	// Initialize axes labels
	pl->SetXLabel("T / K");
	if (isNormed) {
		pl->SetYLabel(wxT("H(T)-H(T0)/(T-T0) / J/mol/K"));
	} else {
		pl->SetYLabel(wxT("H(T)-H(T0) / kJ/mol"));
	}
	// Draw data series and find maxT value
	double maxT = -100;
	std::vector <std::string> serIDList = lsqCont.GetSeriesIDList(false, DATATYPE_H);
	pl->SetLegend(serIDList, seriesMarkers, seriesMarkers[0]);
	unsigned int numOfMarkers = seriesMarkers.size();
	for (unsigned int i = 0; i < serIDList.size(); i++) {	
		int markerIndex = (i >= numOfMarkers) ? (numOfMarkers - 1) : i;
		std::vector <double> x1raw = lsqCont.GetX1Vector(DATATYPE_H, serIDList[i]);
		std::vector <double> x2raw = lsqCont.GetX2Vector(DATATYPE_H, serIDList[i]);
		std::vector <double> yraw = lsqCont.GetYVector(DATATYPE_H, serIDList[i]);
		std::vector <double> x2, y;
		for (unsigned int j = 0; j < yraw.size(); j++) {
			if (T0 - DT0 <= x1raw[j] && x1raw[j] <= T0 + DT0) { // Filter points not matching our T0
				x2.push_back(x2raw[j]);
				if (x2raw[j] > maxT) { // Find maximal temperature (T2)
					maxT = x2raw[j];
				}
				double yj = (isNormed) ? (yraw[j] / (x2raw[j] - T0)) : (yraw[j] / 1000);
				y.push_back(yj);				
			}
		}
		if (x2.size() > 0) {
			pl->Plot(x2, y, seriesMarkers[markerIndex]);
		}
	}
	// Calculate theoretical curve
	std::vector <double> x2 = lsqCont.GetX2Vector(DATATYPE_H);
	std::vector <double> T, T0vec;
	if (maxT <= T0) {
		maxT = T0 + 10;
	}
	for (double Ti = T0 + 0.001; Ti <= maxT + 0.001; Ti += (maxT - T0) * 0.001) {
		T.push_back(Ti);
		T0vec.push_back(T0);
	}
	std::vector <double> DH = lsqRes.DHFunc(T0vec, T);
	for (unsigned int i = 0; i < DH.size(); i++) {
		DH[i] = (isNormed) ? DH[i] / (T[i] - T0vec[i]) : (DH[i] / 1000);
	}
	// Draw theoretical curve
	pl->Plot(T, DH, PlotMarker(wxColor(0,0,0), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	// Enable/disable legend
	EnableLegend();
	// Initialize axes
	pl->FitAxes();
}

/**
 * \brief Initialize Cp(T)/T 2D plot canvas and controls
 */
CpTDataWindow::CpTDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes)
	: XYPlotWindow(parent, title, lsqCont, lsqRes)
{
	InitPlot();
}

/**
 * \brief  Initialize axes labels, sizes and calculate curves
 */
void CpTDataWindow::InitPlot()
{
	static wxColor colorsCpT[8] = {wxColor(0,0,0), wxColor(0,0,255), wxColor(0,128,0), wxColor(0,128,128),
		wxColor(255,0,0), wxColor(255,0,255), wxColor(128,128,0), wxColor(128,128,128)};

	// Initialize axes labels
	pl->SetXLabel(wxT("T / K"));
	pl->SetYLabel(wxT("Cp/T / J/(mol\u00B7K\u00B7K)")); // \cdot
	// Draw data series
	std::vector <std::string> serIDList = lsqCont.GetSeriesIDList(false, DATATYPE_CP);
	pl->SetLegend(serIDList, seriesMarkers, seriesMarkers[0]);
	unsigned int numOfMarkers = seriesMarkers.size();
	for (unsigned int i = 0; i < serIDList.size(); i++) {
		int markerIndex = (i >= numOfMarkers) ? (numOfMarkers - 1) : i;
		std::vector <double> x = lsqCont.GetXVector(DATATYPE_CP, serIDList[i]);
		std::vector <double> y = lsqCont.GetYVector(DATATYPE_CP, serIDList[i]);
		for (unsigned int j = 0; j < x.size(); j++) {
			y[j] /= x[j];
		}
		pl->Plot(x, y, seriesMarkers[markerIndex]);
	}
	// Draw theoretical curve
	// Add to graph
	std::vector <double> x = lsqCont.GetXVector(DATATYPE_CP);
	std::vector <double> y = lsqCont.GetYVector(DATATYPE_CP);
	std::vector <double> T;
	double minT = 1, maxT = 990;
	if (x.size() >= 2) {
		minT = *std::min_element(x.begin(), x.end());
		maxT = *std::max_element(x.begin(), x.end());
	}
	for (double Ti = minT; Ti <= maxT + 1e-10; Ti += (maxT - minT) * 0.001) {
		T.push_back(Ti);		
	}
	std::vector <double> CpT = lsqRes.CpFunc(T);
	for (unsigned int i = 0; i < CpT.size(); i++) {
		CpT[i] /= T[i];
	}
	pl->Plot(T, CpT, PlotMarker(wxColor(0,0,0), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	// Drawing separate terms
	int nterms = lsqRes.GetNumOfParams() / 2;
	std::pair < std::vector <double>, std::vector <double> > pairAlphaTheta = lsqRes.GetAlphaTheta();
	for (int i = 0; i < nterms; i++) {
		TDPEFunc ff;
		ff.AddTerm(pairAlphaTheta.first[i], pairAlphaTheta.second[i]); // alpha & theta
		std::vector <double> CpTi = ff.CpFunc(T);
		for (unsigned int j = 0; j < CpTi.size(); j++) {
			CpTi[j] /= T[j];
		}
		wxColor color = (i <= 7) ? colorsCpT[i] : wxColor(128,0,0);

		pl->Plot(T, CpTi, PlotMarker(color, wxColor(255,0,0), LINE_SHORT_DASH, MARKER_NONE));		
	}
	// Initialize axes
	pl->FitAxes();
}

/**
 * \brief Saves user data set, data type and initializes the 2D plot with residuals (absolute or relative)
 * \param parent Parent window
 * \param title Window title
 * \param lsqCont LSQContext class example (with user data)
 * \param lsqRes LSQResult class example (with user model)
 * \param dataType Selects data type that will be used on graph
 * \param isRel true/false - show relative/absolute deviations on the plot
 */
ResDataWindow::ResDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes, TDDataType dataType, bool isRel)
	: XYPlotWindow(parent, title, lsqCont, lsqRes)
{
	this->dataType = dataType;
	this->isRel = isRel;

	InitPlot();
}

/**
 * \brief  Initialize axes labels, sizes and calculate curves
 */
void ResDataWindow::InitPlot()
{
	// Initialize axes labels
	if (dataType == DATATYPE_H) {
		pl->SetXLabel("T2, K");
	} else {
		pl->SetXLabel("T, K");
	}
	if (isRel) {
		switch (dataType) {
		case DATATYPE_CP:
			pl->SetYLabel(wxT("\u03B5Cp, %")); // \epsilon
			break;
		case DATATYPE_H:
			pl->SetYLabel(wxT("\u03B5(H(T2)-H(T1)), %"));
			break;
		case DATATYPE_S:
			pl->SetYLabel(wxT("\u03B5(S(T)-S(0)), %"));
			break;
		default: // TODO: SOME ERROR MESSAGE
			break;
		}
	} else {
		switch (dataType) {
		case DATATYPE_CP:		
			pl->SetYLabel(wxT("\u03B4Cp / J/(mol\u00B7K)")); // \delta and \cdot
			break;
		case DATATYPE_H:
			pl->SetYLabel(wxT("\u03B4(H(T2)-H(T1)) / J/mol"));
			break;
		case DATATYPE_S:
			pl->SetYLabel(wxT("\u03B4(S(T)-S(0)) / J/(mol\u00B7K)"));
			break;
		default: // TODO: SOME ERROR MESSAGE
			break;
		}
	}
	// Get data
	std::vector <std::string> serIDList = lsqCont.GetSeriesIDList(false, dataType);
	pl->SetLegend(serIDList, seriesMarkers, seriesMarkers[0]);
	unsigned int numOfMarkers = seriesMarkers.size();
	
	for (unsigned int i = 0; i < serIDList.size(); i++) {	
		int markerIndex = (i >= numOfMarkers) ? (numOfMarkers - 1) : i;
		std::vector <double> T = lsqCont.GetXVector(dataType, serIDList[i]);
		std::vector <double> T1 = lsqCont.GetX1Vector(dataType, serIDList[i]);
		std::vector <double> T2 = lsqCont.GetX2Vector(dataType, serIDList[i]);
		std::vector <double> ycalc, yobs = lsqCont.GetYVector(dataType, serIDList[i]);

		if (dataType == DATATYPE_H) {
			ycalc = lsqRes.DHFunc(T1, T2);
			T = T2;
		} else {
			ycalc = lsqRes.YFunc(dataType, T);
		}
		// Calculate uncertainties
		std::vector <double> d;
		for (unsigned int i = 0; i < T.size(); i++) {
			if (isRel) { // Relative deviation
				d.push_back((yobs[i] - ycalc[i]) / yobs[i] * 100.0);
			} else { // Absolute deviation
				d.push_back(yobs[i] - ycalc[i]);
			}
		}
		// Add points to the plot
		pl->Plot(T, d, seriesMarkers[markerIndex]);
	}
	// Initialize axes	
	pl->FitAxes();
}
